package ru.dbotthepony.mc.otm.storage;

import net.minecraft.MethodsReturnNonnullByDefault;
import net.minecraft.core.BlockPos;
import net.minecraft.world.level.Level;
import net.minecraftforge.common.util.LazyOptional;
import ru.dbotthepony.mc.otm.OverdriveThatMatters;
import ru.dbotthepony.mc.otm.capability.IStorageGridCell;

import javax.annotation.ParametersAreNonnullByDefault;
import java.util.*;

@MethodsReturnNonnullByDefault
@ParametersAreNonnullByDefault
public class StorageGrid {
	private final HashSet<IStorageGridCell> cells = new HashSet<>();

	private final HashMap<Class<? extends IStorageStack>, VirtualComponent<? extends IStorageStack>> virtual_components = new HashMap<>();

	public StorageGrid() {}

	public <T extends IStorageStack> T insertObject(Class<T> type, T object, boolean simulate) {
		return getVirtualComponent(type).insertStack(object, simulate);
	}

	/**
	 * @param type Class, representing identity of stored object
	 * @param <T> The identity itself
	 * @return a virtual IStorageComponent<T>, or null if storage grid contain no view for provided type
	 */
	@SuppressWarnings("unchecked")
	public <T extends IStorageStack> VirtualComponent<T> getVirtualComponent(final Class<T> type) {
		return (VirtualComponent<T>) virtual_components.computeIfAbsent(type, (k) -> new VirtualComponent<>(type));
	}

	public int size() {
		return cells.size();
	}

	public <T extends IStorageStack> void add(IStorageIdentity<T> identity) {
		getVirtualComponent(identity.storageIdentity()).add(identity);
	}

	public boolean add(LazyOptional<IStorageGridCell> cell) {
		if (!cell.isPresent())
			return false;

		var resolved = cell.resolve();

		if (resolved.isPresent()) {
			final var value = resolved.get();

			if (add(value)) {
				cell.addListener((lazy) -> remove(value));
				return true;
			}
		}

		return false;
	}

	public boolean add(IStorageGridCell cell) {
		if (cells.add(cell)) {
			cell.setStorageGrid(this);

			for (var identity : cell.getComponents()) {
				add(identity);
			}

			return true;
		}

		return false;
	}

	public <T extends IStorageStack> void remove(IStorageIdentity<T> identity) {
		getVirtualComponent(identity.storageIdentity()).remove(identity);
	}

	public boolean remove(IStorageGridCell cell) {
		if (cells.remove(cell)) {
			cell.setStorageGrid(null);

			for (var identity : cell.getComponents()) {
				remove(identity);
			}

			return true;
		}

		return false;
	}

	public void detach(IStorageGridCell cell) {
		if (cells.contains(cell)) {
			for (var identity : cell.getComponents()) {
				remove(identity);
			}
		}
	}

	public void attach(IStorageGridCell cell) {
		if (cells.contains(cell)) {
			for (var identity : cell.getComponents()) {
				add(identity);
			}
		}
	}

	public StorageGrid merge(StorageGrid other) {
		if (other == this)
			return this;

		if (size() < other.size()) {
			var copy = List.copyOf(cells);

			for (var cell : copy) {
				remove(cell);
				other.add(cell);
			}

			return other;
		} else {
			var copy = List.copyOf(other.cells);

			for (var cell : copy) {
				other.remove(cell);
				add(cell);
			}

			return this;
		}
	}

	public static void scheduleDiscoverNeighbours(IStorageGridCell cell, Level level, BlockPos pos) {
		OverdriveThatMatters.tickUntil(level, () -> !cell.isValidStorageCell() || cell.connectOrCreateStorageGrid(level, pos, true));
	}
}
